

;*****************************************************************************

;             Sample I2C Slave Routines for the 8XC751 and 8XC752

; This program demonstrates I2C slave functions for the 8XC751 and 8XC752 
; microcontrollers. The program uses separate transmit and receive data 
; buffers that are each eight bytes deep. The sample main program 
; copies received data to the transmit buffer such that transmitted data can 
; be read back by a bus master. Buffer addresses 0 and 1 are mapped to port 1 
; and 3 respectively, such that an I2C write will affect the port outputs, and 
; an I2C read will return port pin data. The 751 will accept only eight data 
; bytes in any one I2C transmission, additional bytes will not be 
; acknowledged. Similarly, only eight data bytes may be read from the 751 in 
; any one I2C transmission. This program does not support subaddressing for 
; buffer access. 

;*****************************************************************************


$TITLE(8XC751 I2C Slave Routines)
$DATE(11/23/92)
$MOD752
$DEBUG


;*****************************************************************************

; Value definitions.

CTVAL      EQU     02h                 ; CT1, CT0 bit values for I2C.
MaxBytes   EQU     8                   ; Maximum # of bytes to be sent or 
                                       ;   received.

; Masks for I2CFG bits.

BTIR       EQU     10h                 ; Mask for TIRUN bit.
BMRQ       EQU     40h                 ; Mask for MASTRQ bit.


; Masks for I2CON bits.

BCXA       EQU     80h                 ; Mask for CXA bit.
BIDLE      EQU     40h                 ; Mask for IDLE bit.
BCDR       EQU     20h                 ; Mask for CDR bit.
BCARL      EQU     10h                 ; Mask for CARL bit.
BCSTR      EQU     08h                 ; Mask for CSTR bit.
BCSTP      EQU     04h                 ; Mask for CSTP bit.
BXSTR      EQU     02h                 ; Mask for XSTR bit.
BXSTP      EQU     01h                 ; Mask for XSTP bit.


; RAM locations used by I2C routines.

RcvDat     DATA    10h                 ; I2C receive data buffer (8 bytes).
                                       ;   addresses 10h through 17h.

XmtDat     DATA    18h                 ; I2C transmit data buffer (8 bytes).
                                       ;   addresses 18h through 1Fh.

Flags      DATA    20h                 ; I2C software status flags.
NoAck      BIT     Flags.7             ; Holds negative acknowledge flag.
DatFlag    BIT     Flags.6             ; Tells whether an I2C write operation 
                                       ;   has occurred.

BitCnt     DATA    21h                 ; I2C bit counter.
ByteCnt    DATA    22h                 ; Send/receive byte counter.
TDAT       DATA    23h                 ; Temporary holding register.
MyAddr     DATA    24h                 ; Holds address of THIS slave.

AdrRcvd    DATA    25h                 ; Holds received slave address + R/W.
RWFlag     BIT     AdrRcvd.0           ; Slave read/write flag.


;*****************************************************************************
;                                Begin Code
;*****************************************************************************

; Reset and interrupt vectors.

           AJMP    Reset               ; Reset vector at address 0.


; A timer I timeout usually indicates a 'hung' bus.

           ORG     1Bh                 ; Timer I (I2C timeout) interrupt.
           AJMP    TimerI


; I2C interrupt is used to detect a start while the slave is idle.

           ORG     23h                 ; I2C interrupt.
           PUSH    PSW                 ; Save status.
           PUSH    ACC                 ; Save accumulator.
           CLR     ES                  ; Disable I2C interrupt.
           ACALL   ClrInt              ; Re-enable interrupts.


;*****************************************************************************
;                    Main Transmit and Receive Routines
;*****************************************************************************

Slave:     MOV     I2CON,#BCARL+BCSTP+BCSTR+BCXA ; Clear start status.
           JNB     ATN,$               ; Wait for next data bit.
           MOV     BitCnt,#7           ; Set bit count.

           ACALL   RcvB2               ; Get remainder of slave address.
           MOV     AdrRcvd,A           ; Save received address + R/W bit.
           CLR     ACC.0
           CJNE    A,MyAddr,GoIdle     ; Enter idle mode if not our address.

           JB      RWFlag,Read         ; Read or Write?
           MOV     R0,#RcvDat          ; Set up receive buffer pointer.
           MOV     ByteCnt,#MaxBytes   ; Max 4 bytes can be received.

WrtLoop:   ACALL   SendAck             ; Send acknowledge.
           ACALL   RcvByte             ; Get data byte from master. 
           JNB     DRDY,WLEx           ; Must be end of frame?
           MOV     @R0,A               ; Save data.
           INC     R0                  ; Advance buffer pointer.
           DJNZ    ByteCnt,WrtLoop     ; Back to receive if buffer not full.
           ACALL   SendAck             ; Send acknowledge.
           ACALL   RcvByte             ; Get, but do not store add'l data.
           MOV     I2DAT,#80h          ; Send negative acknowledge.
           JNB     ATN,$               ; Wait for acknowledge sent.
WLEx:      SETB    DatFlag             ; Flag main that data has been received.
           SJMP    MsgEnd              ; Buffer full, enter idle mode.


Read:      MOV     R0,#XmtDat          ; Set up transmit buffer pointer.
           MOV     ByteCnt,#MaxBytes   ; Max bytes to be sent.
           ACALL   SendAck             ; Send address acknowledge.

RdLoop:    MOV     A,@R0               ; Get data byte from buffer.
           CJNE    R0,#XmtDat,RdL1     ; Return port 1 value instead of buffer
           MOV     A,P1                ;   data if this is buffer address 0.
RdL1:      CJNE    R0,#XmtDat+1,RdL2   ; Return port 3 value instead of buffer
           MOV     A,P3                ;   data if this is buffer address 1.

RdL2:      INC     R0                  ; Advance buffer pointer.
           ACALL   XmitByte            ; Send data byte.
           JB      NoAck,RLEx          ; Exit if NAK.
           DJNZ    ByteCnt,RdLoop      ; Back if more data requested & avail.
RLEx:      SJMP    MsgEnd              ; Done, enter idle mode.

MsgEnd:    JNB     ATN,$               ; Wait for stop or repeated start.
           JB      STR,Slave           ; If repeated start, go to slave mode,
                                       ;   else enter idle mode.

GoIdle:    MOV     I2CON,#BCSTP+BCXA+BCDR+BCARL+BIDLE ; Enter slave idle mode.
           POP     ACC                 ; Restore accumulator.
           POP     PSW                 ; Restore status.
           SETB    ES                  ; Re-enable I2C interrupts.
           RET


;*****************************************************************************
;                               Subroutines
;*****************************************************************************


; Byte transmit routine.
;   Enter with data in ACC.

XmitByte:  MOV     BitCnt,#8           ; Set 8 bits of data count.
XmBit:     MOV     I2DAT,A             ; Send this bit.
           RL      A                   ; Get next bit.
           JNB     ATN,$               ; Wait for bit sent.
           DJNZ    BitCnt,XmBit        ; Repeat until all bits sent.
           MOV     I2CON,#BCDR+BCXA    ; Switch to receive mode.
           JNB     ATN,$               ; Wait for acknowledge bit.
           MOV     Flags,I2DAT         ; Save acknowledge bit.
           RET


; Byte receive routines.
;   SendAck : sends an I2C acknowledge.
;   RcvByte : receives a byte of data.
;   RcvB2   : receives a partial byte of I2C data, used to allow reception of 
;               7 bits of slave address information.
;   Data is returned in the ACC.

SendAck:   MOV     I2DAT,#0            ; Send receive acknowledge.
           JNB     ATN,$               ; Wait for acknowledge sent.
           RET

RcvByte:   MOV     BitCnt,#8           ; Set bit count.
RcvB2:     CLR     A                   ; Init received byte to 0.
RBit:      ORL     A,I2DAT             ; Get bit, clear ATN.
           RL      A                   ; Shift data.
           JNB     ATN,$               ; Wait for next bit.
           JNB     DRDY,RBEx           ; Exit if not a data bit.
           DJNZ    BitCnt,RBit         ; Repeat until 7 bits are in.
           MOV     C,RDAT              ; Get last bit, don't clear ATN.
           RLC     A                   ; Form full data byte.
RBEx:      RET


; Timer I timeout interrupt service routine.

TimerI:    SETB    CLRTI               ; Clear timer I interrupt.
           MOV     I2CFG,#0            ; Turn off I2C.
           MOV     I2CON,#BCXA+BCDR+BCARL+BCSTR+BCSTP ; Reset I2C flags.
           ACALL   ClrInt              ; Clear interrupt pending flag.
           AJMP    Reset               ; Return to mainline.
ClrInt:    RETI



;*****************************************************************************
;                               Main Program
;*****************************************************************************

Reset:     MOV     SP,#2Fh             ; Set stack location.
           MOV     IE,#90h             ; Enable I2C interrupt.

           MOV     R0,#RcvDat          ; Set up pointer to data area.
           MOV     R1,#2*MaxBytes      ; Set up buffer length counter.
RLoop:     MOV     @R0,#0              ; Clear buffer memory.
           INC     R0                  ; Advance to next buffer position.
           DJNZ    R1,RLoop            ; Repeat until done.

           MOV     MyAddr,#40h         ; Set slave address.
           MOV     Flags,#0            ; Clear system flags.
           MOV     I2CFG,#80h+CTVAL    ; Enable slave functions.
           MOV     I2CON,#BIDLE        ; Put slave into idle mode.
           SETB    ETI                 ; Enable timer I interrupts.
           SETB    TIRUN               ; Turn on timer I.


; This sample mainline program copies the first two received bytes to Port 1 
;   and Port 3 whenever there is an I2C write operation. It Also copies the 
;   rest of the input buffer to the output buffer at the same time. 

MainLoop:  JNB     DatFlag,$           ; Wait for data sent from I2C.
           CLR     EA                  ; Turn off interrupts during data move.

           MOV     P1,RcvDat           ; First buffer location goes to port 1.
           MOV     P3,RcvDat+1         ; Second buffer location goes to port 3.

           MOV     R0,#RcvDat          ; Set input buffer start pointer.
           MOV     R1,#XmtDat          ; Set output buffer start pointer.
           MOV     R2,#MaxBytes        ; Set buffer length counter.
ML2:       MOV     A,@R0               ; Get data from input buffer.
           MOV     @R1,A               ; Store data in output buffer.
           INC     R1                  ; Increment input buffer pointer.
           INC     R0                  ; Increment output buffer pointer.
           DJNZ    R2,ML2              ; Repeat until entire buffer is updated.
           CLR     DatFlag             ; Clear I2C transmission flag.

           SETB    EA                  ; Data move done, re-enable interrupts.
           SJMP    MainLoop            ; Wait for next I2C transmission.

           END
